clear all

%% Load Data
maxIter = 3;

lambda = 1;
[Xnode,Xedge,Y,edgeStruct,infoStruct,nRows,nCols] = prepareBinaryDenoising;
nInstances = size(Y,1);
[wN,wE] = UGM_initWeights(infoStruct);
x_init = [wN(:);wE(:)];
funObj = @(w)penalizedL2(w,@UGM_CRFpseudoLoss,lambda,Xnode,Xedge,Y,edgeStruct,infoStruct);
funObjBatch = @(w,batch)penalizedL2(w,@UGM_CRFpseudoLoss,lambda*length(batch)/nInstances,Xnode(batch,:,:),Xedge(batch,:,:),Y(batch,:,:),edgeStruct,infoStruct);

% Show examples of noisy images
figure(1);
for i = 1:4
	subplot(2,2,i);
	imagesc(-reshape(Xnode(i,2,:),nRows,nCols));
	colormap gray
end
suptitle('Examples of Noise-Corrupted Images');
fprintf('(paused)\n');
pause

%% Evaluate with random parameters

figure(2);
fprintf('Testing random parameters...\n');
[w,v] = UGM_initWeights(infoStruct,@randn);
for i = 1:4
	subplot(2,2,i);
	nodePot = UGM_makeCRFNodePotentials(Xnode,w,edgeStruct,infoStruct);
	edgePot = UGM_makeCRFEdgePotentials(Xedge,v,edgeStruct,infoStruct);
	nodeBel = UGM_Infer_LBP(nodePot(:,:,i),edgePot(:,:,:,i),edgeStruct);
	imagesc(-reshape(nodeBel(:,2),nRows,nCols));
	colormap gray
end
suptitle('Loopy BP node marginals with random parameters');
fprintf('(paused)\n');
pause

%% L-BFGS
fprintf('Estimating parameters with deterministic algorithm...\n');
options.maxIter = maxIter;
options.maxFunEvals = maxIter;
options.Display = 'Full';
options.Corr = 10;
x = minFunc(funObj,x_init,options);

figure(3);
[w,v] = UGM_splitWeights(x,infoStruct);
for i = 1:4
	subplot(2,2,i);
	nodePot = UGM_makeCRFNodePotentials(Xnode,w,edgeStruct,infoStruct);
	edgePot = UGM_makeCRFEdgePotentials(Xedge,v,edgeStruct,infoStruct);
	nodeBel = UGM_Infer_LBP(nodePot(:,:,i),edgePot(:,:,:,i),edgeStruct);
	imagesc(-reshape(nodeBel(:,2),nRows,nCols));
	colormap gray
end
suptitle(sprintf('Loopy BP node marginals after %d passes of deterministic algorithm',maxIter));
fprintf('(paused)\n');
pause

%% Stochastic Gradient
fprintf('Estimating parameters with stochastic algorithm...\n');
stepSize = 1e-4;
x = x_init;
for iter = 1:maxIter*nInstances
	i = ceil(rand*nInstances);
	[f,g] = funObjBatch(x,i);
	fprintf('Iter = %d of %d (fsub = %f)\n',iter,maxIter*nInstances,f);
	x = x - stepSize*g;
end

figure(4);
[w,v] = UGM_splitWeights(x,infoStruct);
for i = 1:4
	subplot(2,2,i);
	nodePot = UGM_makeCRFNodePotentials(Xnode,w,edgeStruct,infoStruct);
	edgePot = UGM_makeCRFEdgePotentials(Xedge,v,edgeStruct,infoStruct);
	nodeBel = UGM_Infer_LBP(nodePot(:,:,i),edgePot(:,:,:,i),edgeStruct);
	imagesc(-reshape(nodeBel(:,2),nRows,nCols));
	colormap gray
end
suptitle(sprintf('Loopy BP node marginals after %d passes of stochastic algorithm',maxIter));
fprintf('(paused)\n');
pause

%% Hybrid
fprintf('Estimating parameters with hybrid algorithm...\n');
options.t0 = 1e-8;
options.gamma = 1.1;
options.delta = 1;
x = batchingLBFGS(funObjBatch,x_init,nInstances,options);

figure(5);
[w,v] = UGM_splitWeights(x,infoStruct);
for i = 1:4
	subplot(2,2,i);
	nodePot = UGM_makeCRFNodePotentials(Xnode,w,edgeStruct,infoStruct);
	edgePot = UGM_makeCRFEdgePotentials(Xedge,v,edgeStruct,infoStruct);
	nodeBel = UGM_Infer_LBP(nodePot(:,:,i),edgePot(:,:,:,i),edgeStruct);
	imagesc(-reshape(nodeBel(:,2),nRows,nCols));
	colormap gray
end
suptitle(sprintf('Loopy BP node marginals after %d passes of hybrid algorithm',maxIter));
fprintf('(paused)\n');
pause